import pytest
from dagster_databricks.databricks import DatabricksError, DatabricksJobRunner, DatabricksRunState
from dagster_databricks.types import DatabricksRunLifeCycleState, DatabricksRunResultState

import dagster
from dagster.seven import mock
from dagster.utils.test import create_test_pipeline_execution_context

HOST = 'https://uksouth.azuredatabricks.net'
TOKEN = 'super-secret-token'


@mock.patch('dagster_databricks.databricks.DatabricksClient.submit_run')
def test_databricks_submit_job_existing_cluster(mock_submit_run, databricks_run_config):
    mock_submit_run.return_value = {'run_id': 1}

    runner = DatabricksJobRunner(HOST, TOKEN)
    task = databricks_run_config.pop('task')
    runner.submit_run(databricks_run_config, task)
    mock_submit_run.assert_called_once_with(
        run_name=databricks_run_config['run_name'],
        new_cluster=None,
        existing_cluster_id=databricks_run_config['cluster']['existing'],
        spark_jar_task=task['spark_jar_task'],
        libraries=[
            {'pypi': {'package': 'dagster=={}'.format(dagster.__version__)}},
            {'pypi': {'package': 'dagster_databricks=={}'.format(dagster.__version__)}},
            {'pypi': {'package': 'dagster_pyspark=={}'.format(dagster.__version__)}},
        ],
    )


@mock.patch('dagster_databricks.databricks.DatabricksClient.submit_run')
def test_databricks_submit_job_new_cluster(mock_submit_run, databricks_run_config):
    mock_submit_run.return_value = {'run_id': 1}

    runner = DatabricksJobRunner(HOST, TOKEN)

    NEW_CLUSTER = {
        'size': {'num_workers': 1},
        'spark_version': '6.5.x-scala2.11',
        'nodes': {'node_types': {'node_type_id': 'Standard_DS3_v2'}},
    }
    databricks_run_config['cluster'] = {'new': NEW_CLUSTER}

    task = databricks_run_config.pop('task')
    runner.submit_run(databricks_run_config, task)
    mock_submit_run.assert_called_once_with(
        run_name=databricks_run_config['run_name'],
        new_cluster={
            'num_workers': 1,
            'spark_version': '6.5.x-scala2.11',
            'node_type_id': 'Standard_DS3_v2',
            'custom_tags': [{'key': '__dagster_version', 'value': dagster.__version__}],
        },
        existing_cluster_id=None,
        spark_jar_task=task['spark_jar_task'],
        libraries=[
            {'pypi': {'package': 'dagster=={}'.format(dagster.__version__)}},
            {'pypi': {'package': 'dagster_databricks=={}'.format(dagster.__version__)}},
            {'pypi': {'package': 'dagster_pyspark=={}'.format(dagster.__version__)}},
        ],
    )


@mock.patch('dagster_databricks.databricks.DatabricksClient.submit_run')
def test_databricks_wait_for_run(mock_submit_run, databricks_run_config):
    mock_submit_run.return_value = {'run_id': 1}

    context = create_test_pipeline_execution_context()
    runner = DatabricksJobRunner(HOST, TOKEN, poll_interval_sec=0.01)
    task = databricks_run_config.pop('task')
    databricks_run_id = runner.submit_run(databricks_run_config, task)

    calls = {
        'num_calls': 0,
        'final_state': DatabricksRunState(
            DatabricksRunLifeCycleState.Terminated, DatabricksRunResultState.Success, 'Finished',
        ),
    }

    def new_get_run_state(_run_id):
        calls['num_calls'] += 1

        if calls['num_calls'] == 1:
            return DatabricksRunState(DatabricksRunLifeCycleState.Pending, None, None,)
        elif calls['num_calls'] == 2:
            return DatabricksRunState(DatabricksRunLifeCycleState.Running, None, None,)
        else:
            return calls['final_state']

    with mock.patch.object(runner.client, 'get_run_state', new=new_get_run_state):
        runner.wait_for_run_to_complete(context.log, databricks_run_id)

    calls['num_calls'] = 0
    calls['final_state'] = DatabricksRunState(
        DatabricksRunLifeCycleState.Terminated, DatabricksRunResultState.Failed, 'Failed',
    )
    with pytest.raises(DatabricksError) as exc_info:
        with mock.patch.object(runner.client, 'get_run_state', new=new_get_run_state):
            runner.wait_for_run_to_complete(context.log, databricks_run_id)
    assert 'Run 1 failed with result state' in str(exc_info.value)



from dagster_databricks import DatabricksRunJobSolidDefinition

from dagster import execute_pipeline, pipeline
from dagster.seven import mock

HOST = 'https://uksouth.azuredatabricks.net'
TOKEN = 'super-secret-token'


@mock.patch('dagster_databricks.databricks.DatabricksJobRunner.submit_run')
@mock.patch('dagster_databricks.databricks.DatabricksJobRunner.wait_for_run_to_complete')
def test_run_databricks_job(mock_wait_for_run, mock_submit_run, databricks_run_config):
    @pipeline
    def test_pipe():
        DatabricksRunJobSolidDefinition('test', poll_interval_sec=1)()

    RUN_ID = 1
    mock_submit_run.return_value = RUN_ID

    config = {
        'solids': {
            'test': {
                'config': {
                    'run_config': databricks_run_config,
                    'databricks_host': HOST,
                    'databricks_token': TOKEN,
                }
            }
        }
    }
    result = execute_pipeline(test_pipe, config)
    assert result.success

    assert mock_submit_run.call_count == 1
    task = databricks_run_config.pop('task')
    assert mock_submit_run.call_args[0][0] == databricks_run_config
    assert mock_submit_run.call_args[0][1] == task
    assert mock_wait_for_run.call_count == 1
    assert mock_wait_for_run.call_args[0][1] == RUN_ID



'''The main entrypoint for solids executed using the DatabricksPySparkStepLauncher.

This script is launched on Databricks using a `spark_python_task` and is passed the following
parameters:

- the DBFS path to the pickled `step_run_ref` file
- the DBFS path to the zipped pipeline
- paths to any other zipped packages which have been uploaded to DBFS.
'''

import os
import pickle
import site
import sys
import tempfile
import zipfile

from dagster import check
from dagster.core.execution.plan.external_step import PICKLED_EVENTS_FILE_NAME, run_step_from_ref
from dagster.serdes import serialize_value

# This won't be set in Databricks but is needed to be non-None for the
# Dagster step to run.
if 'DATABRICKS_TOKEN' not in os.environ:
    os.environ['DATABRICKS_TOKEN'] = ''


def setup_s3_storage(storage):
    '''Obtain AWS credentials from Databricks secrets and export so both Spark and boto can use them.
    '''

    scope = storage['secret_scope']

    # dbutils is globally defined in the Databricks runtime
    access_key = dbutils.secrets.get(  # noqa  # pylint: disable=undefined-variable
        scope=scope, key=storage['access_key_key']
    )
    secret_key = dbutils.secrets.get(  # noqa  # pylint: disable=undefined-variable
        scope=scope, key=storage['secret_key_key']
    )

    # Spark APIs will use this.
    # See https://docs.databricks.com/data/data-sources/aws/amazon-s3.html#alternative-1-set-aws-keys-in-the-spark-context.
    # sc is globally defined in the Databricks runtime and points to the Spark context
    sc._jsc.hadoopConfiguration().set(  # noqa  # pylint: disable=undefined-variable,protected-access
        'fs.s3n.awsAccessKeyId', access_key
    )
    sc._jsc.hadoopConfiguration().set(  # noqa  # pylint: disable=undefined-variable,protected-access
        'fs.s3n.awsSecretAccessKey', secret_key
    )

    # Boto will use these.
    os.environ['AWS_ACCESS_KEY_ID'] = access_key
    os.environ['AWS_SECRET_ACCESS_KEY'] = secret_key


def setup_adls2_storage(storage):
    '''Obtain an Azure Storage Account key from Databricks secrets and export so Spark can use it.
    '''
    # dbutils is globally defined in the Databricks runtime
    storage_account_key = dbutils.secrets.get(  # noqa  # pylint: disable=undefined-variable
        scope=storage['secret_scope'], key=storage['storage_account_key_key']
    )
    # Spark APIs will use this.
    # See https://docs.microsoft.com/en-gb/azure/databricks/data/data-sources/azure/azure-datalake-gen2#--access-directly-using-the-storage-account-access-key
    # sc is globally defined in the Databricks runtime and points to the Spark context
    sc._jsc.hadoopConfiguration().set(  # noqa  # pylint: disable=undefined-variable,protected-access
        'fs.azure.account.key.{}.dfs.core.windows.net'.format(storage['storage_account_name']),
        storage_account_key,
    )


def setup_storage(step_run_ref):
    '''Setup any storage required by the run.

    At least one of S3 or ADLS2 storage should be provided in config, so that the run can
    save intermediate files to a location accessible by the original process which launched
    the job.

    This requires modifying the 'sc' global which isn't great.
    https://github.com/dagster-io/dagster/issues/2492 tracks a better solution
    '''
    root_storage = step_run_ref.run_config['storage']
    check.invariant(
        's3' in root_storage or 'adls2' in root_storage, 'No compatible storage found in config'
    )

    storage = step_run_ref.run_config['resources']['pyspark_step_launcher']['config']['storage']
    check.invariant(
        len(storage) == 1, 'No valid storage credentials found in pyspark_step_launcher config'
    )
    check.invariant(
        list(storage)[0] == list(root_storage)[0],
        "Storage credentials in step launcher config don't match root storage",
    )

    if 's3' in storage:
        setup_s3_storage(storage['s3'])
    elif 'adls2' in storage:
        setup_adls2_storage(storage['adls2'])
    else:
        raise Exception('No valid storage found!')


def main(step_run_ref_filepath, pipeline_zip):
    # Extract any zip files to a temporary directory and add that temporary directory
    # to the site path so the contained files can be imported.
    #
    # We can't rely on pip or other packaging tools because the zipped files might not
    # even be Python packages.
    with tempfile.TemporaryDirectory() as tmp:

        with zipfile.ZipFile(pipeline_zip) as zf:
            zf.extractall(tmp)
        site.addsitedir(tmp)

        # We can use regular local filesystem APIs to access DBFS inside the Databricks runtime.
        with open(step_run_ref_filepath, 'rb') as handle:
            step_run_ref = pickle.load(handle)

        setup_storage(step_run_ref)

        events = list(run_step_from_ref(step_run_ref))

    events_filepath = os.path.dirname(step_run_ref_filepath) + '/' + PICKLED_EVENTS_FILE_NAME
    with open(events_filepath, 'wb') as handle:
        pickle.dump(serialize_value(events), handle)


if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2])



# -*- coding: utf-8 -*-
'''Fields here are taken from the Databricks API docs.

Most are left the same, but some have been modified to better express
the requirements.

See:
- https://docs.databricks.com/dev-tools/api/latest/jobs.html
- https://docs.databricks.com/dev-tools/api/latest/clusters.html
- https://docs.databricks.com/dev-tools/api/latest/libraries.html
'''
from dagster import Bool, Field, Int, Permissive, Selector, Shape, String


def _define_autoscale():
    return Field(
        Shape(
            fields={
                'min_workers': Field(
                    Int,
                    description='The minimum number of workers to which the cluster can scale down '
                    'when underutilized. It is also the initial number of workers the cluster '
                    'will have after creation.',
                ),
                'max_workers': Field(
                    Int,
                    description='The maximum number of workers to which the cluster can scale up '
                    'when overloaded. max_workers must be strictly greater than min_workers.',
                ),
            }
        )
    )


def _define_size():
    num_workers = Field(
        Int,
        description='If num_workers, number of worker nodes that this cluster should have. '
        'A cluster has one Spark Driver and num_workers Executors for a total of '
        'num_workers + 1 Spark nodes.',
        is_required=True,
    )
    return Selector({'autoscale': _define_autoscale(), 'num_workers': num_workers})


def _define_custom_tags():
    key = Field(
        String,
        description='The key of the tag. The key length must be between 1 and 127 UTF-8 '
        'characters, inclusive. For a list of all restrictions, see AWS Tag Restrictions: '
        'https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/Using_Tags.html#tag-restrictions',
        is_required=True,
    )
    value = Field(
        String,
        description='The value of the tag. The value length must be less than or equal to '
        '255 UTF-8 characters. For a list of all restrictions, see AWS Tag Restrictions: '
        'https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/Using_Tags.html#tag-restrictions',
        is_required=True,
    )
    return Field(
        [Shape(fields={'key': key, 'value': value})],
        description='Additional tags for cluster resources. Databricks tags all cluster resources '
        '(e.g., AWS instances and EBS volumes) with these tags in addition to default_tags. Note: '
        '- Tags are not supported on legacy node types such as compute-optimized and '
        'memory-optimized '
        '- Databricks allows at most 45 custom tags'
        'More restrictions may apply if using Azure Databricks; refer to the official docs '
        'for further details.',
        is_required=False,
    )


def _define_dbfs_storage_info():
    destination = Field(String, description='DBFS destination, e.g. dbfs:/my/path')
    return Field(Shape(fields={'destination': destination}), description='DBFS storage information')


def _define_s3_storage_info():
    destination = Field(
        String,
        description='S3 destination, e.g. s3://my-bucket/some-prefix. '
        'You must configure the cluster with an instance profile and the instance profile '
        'must have write access to the destination. You cannot use AWS keys.',
    )
    region = Field(
        String,
        description='S3 region, e.g. us-west-2. Either region or endpoint must be set. '
        'If both are set, endpoint is used.',
    )
    endpoint = Field(
        String,
        description='S3 endpoint, e.g. https://s3-us-west-2.amazonaws.com. '
        'Either region or endpoint must be set. If both are set, endpoint is used.',
    )
    enable_encryption = Field(
        Bool,
        description='(Optional) Enable server side encryption, false by default.',
        is_required=False,
    )
    encryption_type = Field(
        String,
        description='(Optional) The encryption type, it could be sse-s3 or sse-kms. '
        'It is used only when encryption is enabled and the default type is sse-s3.',
        is_required=False,
    )
    kms_key = Field(
        String,
        description='(Optional) KMS key used if encryption is enabled and encryption type is set '
        'to sse-kms.',
        is_required=False,
    )
    canned_acl = Field(
        String,
        description='(Optional) Set canned access control list, e.g. bucket-owner-full-control.'
        'If canned_acl is set, the cluster instance profile must have s3:PutObjectAcl permission '
        'on the destination bucket and prefix. '
        'The full list of possible canned ACLs can be found at '
        'https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl. '
        'By default only the object owner gets full control. '
        'If you are using cross account role for writing data, you may want to set '
        'bucket-owner-full-control to make bucket owner able to read the logs.',
        is_required=False,
    )
    return Field(
        Shape(
            fields={
                'destination': destination,
                'region': region,
                'endpoint': endpoint,
                'enable_encryption': enable_encryption,
                'encryption_type': encryption_type,
                'kms_key': kms_key,
                'canned_acl': canned_acl,
            }
        ),
        description='S3 storage information',
    )


def _define_cluster_log_conf():
    return Field(
        Selector({'dbfs': _define_dbfs_storage_info(), 's3': _define_s3_storage_info()}),
        description='Recommended! The configuration for delivering Spark logs to a long-term '
        'storage destination. Only one destination can be specified for one cluster. If the conf '
        'is given, the logs will be delivered to the destination every 5 mins. '
        'The destination of driver logs is <destination>/<cluster-id>/driver, while the '
        'destination of executor logs is <destination>/<cluster-id>/executor.',
        is_required=False,
    )


def _define_init_script():
    return Selector({'dbfs': _define_dbfs_storage_info(), 's3': _define_s3_storage_info()})


def _define_node_types():
    node_type_id = Field(
        String,
        description='This field encodes, through a single value, the resources available to each '
        'of the Spark nodes in this cluster. For example, the Spark nodes can be provisioned '
        'and optimized for memory or compute intensive workloads. '
        'A list of available node types can be retrieved by using the List node types API '
        'call. This field is required.',
        is_required=True,
    )

    driver_node_type_id = Field(
        String,
        description='The node type of the Spark driver. '
        'This field is optional; if unset, the driver node type is set as the '
        'same value as node_type_id defined above.',
        is_required=False,
    )

    return Field(
        Shape(fields={'node_type_id': node_type_id, 'driver_node_type_id': driver_node_type_id})
    )


def _define_nodes():
    instance_pool_id = Field(
        String,
        description='The optional ID of the instance pool to which the cluster belongs. '
        'Refer to the Instance Pools API for details.',
        is_required=False,
    )

    return Field(
        Selector({'node_types': _define_node_types(), 'instance_pool_id': instance_pool_id}),
        description='The nodes used in the cluster. Either the node types or an instance pool '
        'can be specified.',
        is_required=True,
    )


def _define_new_cluster():
    spark_version = Field(
        String,
        description='The Spark version of the cluster. '
        'A list of available Spark versions can be retrieved by using the '
        'Runtime versions API call. This field is required.',
        is_required=True,
    )

    spark_conf = Field(
        Permissive(),
        description='An object containing a set of optional, user-specified Spark configuration '
        'key-value pairs. You can also pass in a string of extra JVM options to the driver and the '
        'executors via spark.driver.extraJavaOptions and spark.executor.extraJavaOptions '
        'respectively. '
        'Example Spark confs: '
        '{"spark.speculation": true, "spark.streaming.ui.retainedBatches": 5} or '
        '{"spark.driver.extraJavaOptions": "-verbose:gc -XX:+PrintGCDetails"}',
        is_required=False,
    )

    ssh_public_keys = Field(
        [String],
        description='SSH public key contents that will be added to each Spark node in this '
        'cluster. The corresponding private keys can be used to login with the user name ubuntu '
        'on port 2200. Up to 10 keys can be specified.',
        is_required=False,
    )

    init_scripts = Field(
        [_define_init_script()],
        description='The configuration for storing init scripts. Any number of scripts can be '
        'specified. The scripts are executed sequentially in the order provided. '
        'If cluster_log_conf is specified, init script logs are sent to '
        '<destination>/<cluster-id>/init_scripts.',
        is_required=False,
    )

    spark_env_vars = Field(
        Permissive(),
        description='An object containing a set of optional, user-specified environment variable '
        'key-value pairs. Key-value pair of the form (X,Y) are exported as is (i.e., export X="Y") '
        'while launching the driver and workers. '
        'To specify an additional set of SPARK_DAEMON_JAVA_OPTS, we recommend appending them '
        'to $SPARK_DAEMON_JAVA_OPTS as shown in the example below. '
        'This ensures that all default Databricks managed environmental variables are '
        'included as well. '
        'Example Spark environment variables: '
        '{"SPARK_WORKER_MEMORY": "28000m", "SPARK_LOCAL_DIRS": "/local_disk0"} or '
        '{"SPARK_DAEMON_JAVA_OPTS": "$SPARK_DAEMON_JAVA_OPTS -Dspark.shuffle.service.enabled=true"}',
        is_required=False,
    )

    enable_elastic_disk = Field(
        Bool,
        description='Autoscaling Local Storage: when enabled, this cluster dynamically acquires '
        'attitional disk space when its Spark workers are running low on disk space. '
        'This feature requires specific AWS permissions to function correctly - '
        'refer to https://docs.databricks.com/clusters/configure.html#autoscaling-local-storage '
        'for details.',
        is_required=False,
    )

    return Field(
        Shape(
            fields={
                'size': _define_size(),
                'spark_version': spark_version,
                'spark_conf': spark_conf,
                'nodes': _define_nodes(),
                'ssh_public_keys': ssh_public_keys,
                'custom_tags': _define_custom_tags(),
                'cluster_log_conf': _define_cluster_log_conf(),
                'init_scripts': init_scripts,
                'spark_env_vars': spark_env_vars,
                'enable_elastic_disk': enable_elastic_disk,
            }
        )
    )


def _define_cluster():
    existing_cluster_id = Field(
        String,
        description='The ID of an existing cluster that will be used for all runs '
        'of this job. When running jobs on an existing cluster, you may '
        'need to manually restart the cluster if it stops responding. '
        'Databricks suggests running jobs on new clusters for '
        'greater reliability.',
        is_required=True,
    )
    return Selector({'new': _define_new_cluster(), 'existing': existing_cluster_id})


def _define_pypi_library():
    package = Field(
        String,
        description='The name of the PyPI package to install. '
        'An optional exact version specification is also supported. '
        'Examples: '
        '- simplejson '
        '- simplejson==3.8.0',
        is_required=True,
    )
    repo = Field(
        String,
        description='The repository where the package can be found. '
        'If not specified, the default pip index is used.',
        is_required=False,
    )
    return Field(
        Shape(fields={'package': package, 'repo': repo}),
        description='Specification of a PyPI library to be installed.',
    )


def _define_maven_library():
    coordinates = Field(
        String,
        description='Gradle-style Maven coordinates. For example: org.jsoup:jsoup:1.7.2. '
        'This field is required.',
        is_required=True,
    )
    repo = Field(
        String,
        description='Maven repo to install the Maven package from. '
        'If omitted, both Maven Central Repository and Spark Packages are searched.',
        is_required=False,
    )
    exclusions = Field(
        [String],
        description='List of dependences to exclude. For example: '
        '["slf4j:slf4j", "*:hadoop-client"]. '
        'Maven dependency exclusions: '
        'https://maven.apache.org/guides/introduction/introduction-to-optional-and-excludes-dependencies.html.',
        is_required=False,
    )
    return Field(
        Shape(fields={'coordinates': coordinates, 'repo': repo, 'exclusions': exclusions}),
        description='Specification of a Maven library to be installed.',
    )


def _define_cran_library():
    package = Field(
        String,
        description='The name of the CRAN package to install. This field is required.',
        is_required=True,
    )
    repo = Field(
        String,
        description='The repository where the package can be found. '
        'If not specified, the default CRAN repo is used.',
        is_required=False,
    )
    return Field(
        Shape(fields={'package': package, 'repo': repo}),
        description='Specification of a CRAN library to be installed.',
    )


def _define_libraries():
    jar = Field(
        String,
        description='URI of the JAR to be installed. DBFS and S3 URIs are supported. '
        'For example: { "jar": "dbfs:/mnt/databricks/library.jar" } or '
        '{ "jar": "s3://my-bucket/library.jar" }. If S3 is used, make sure the cluster has '
        'read access on the library. '
        'You may need to launch the cluster with an instance profile to access the S3 URI.',
    )
    egg = Field(
        String,
        description='URI of the egg to be installed. DBFS and S3 URIs are supported. '
        'For example: { "egg": "dbfs:/my/egg" } or { "egg": "s3://my-bucket/egg" }. '
        'If S3 is used, make sure the cluster has read access on the library. '
        'You may need to launch the cluster with an instance profile to access the S3 URI. ',
    )
    whl = Field(
        String,
        description='URI of the wheel or zipped wheels to be installed. DBFS and S3 URIs are '
        'supported. For example: { "whl": "dbfs:/my/whl" } or { "whl": "s3://my-bucket/whl" }.'
        'If S3 is used, make sure the cluster has read access on the library. '
        'You may need to launch the cluster with an instance profile to access the S3 URI. '
        'Also the wheel file name needs to use the correct convention '
        '(https://www.python.org/dev/peps/pep-0427/#file-format). '
        'If zipped wheels are to be installed, the file name suffix should be .wheelhouse.zip.',
    )
    return Field(
        [
            Selector(
                {
                    'jar': jar,
                    'egg': egg,
                    'whl': whl,
                    'pypi': _define_pypi_library(),
                    'maven': _define_maven_library(),
                    'cran': _define_cran_library(),
                }
            )
        ],
        description='An optional list of libraries to be installed on the cluster that will '
        'execute the job. By default dagster, dagster-databricks and dagster-pyspark libraries '
        'will be included.',
        is_required=False,
    )


def _define_submit_run_fields():
    run_name = Field(
        String,
        description='An optional name for the run. The default value is Untitled',
        is_required=False,
    )
    timeout_seconds = Field(
        Int,
        description='An optional timeout applied to each run of this job. '
        'The default behavior is to have no timeout.',
        is_required=False,
    )
    idempotency_token = Field(
        String,
        description='An optional token that can be used to guarantee the idempotency of job run requests.'
        'If an active run with the provided token already exists, the request will not create '
        'a new run, but will return the ID of the existing run instead. '
        'If you specify the idempotency token, upon failure you can retry until the request '
        'succeeds. Databricks guarantees that exactly one run will be launched with that '
        'idempotency token. '
        'This token should have at most 64 characters.',
        is_required=False,
    )
    return {
        'cluster': _define_cluster(),
        'run_name': run_name,
        'libraries': _define_libraries(),
        'timeout_seconds': timeout_seconds,
        'idempotency_token': idempotency_token,
    }


def _define_notebook_task():
    notebook_path = Field(
        String,
        description='The absolute path of the notebook to be run in the Databricks Workspace. '
        'This path must begin with a slash. This field is required.',
        is_required=True,
    )
    base_parameters = Field(
        Permissive(),
        description='Base parameters to be used for each run of this job. '
        "If the notebook takes a parameter that is not specified in the job's base_parameters "
        'or the run-now override parameters, the default value from the notebook will be used. '
        'Retrieve these parameters in a notebook by using dbutils.widgets.get().',
        is_required=False,
    )
    return Field(Shape(fields={'notebook_path': notebook_path, 'base_parameters': base_parameters}))


def _define_spark_jar_task():
    main_class_name = Field(
        String,
        description='The full name of the class containing the main method to be executed. '
        'This class must be contained in a JAR provided as a library. '
        'The code should use SparkContext.getOrCreate to obtain a Spark context; '
        'otherwise, runs of the job will fail.',
        is_required=True,
    )
    parameters = Field(
        [String],
        description='Parameters that will be passed to the main method.',
        is_required=False,
        default_value=[],
    )
    return Field(Shape(fields={'main_class_name': main_class_name, 'parameters': parameters}))


def _define_spark_python_task():
    python_file = Field(
        String,
        description='The URI of the Python file to be executed. DBFS and S3 paths are supported.'
        'This field is required.',
        is_required=True,
    )
    parameters = Field(
        [String],
        description='Command line parameters that will be passed to the Python file.',
        is_required=False,
        default_value=[],
    )
    return Field(Shape(fields={'python_file': python_file, 'parameters': parameters}))


def _define_spark_submit_task():
    parameters = Field(
        [String], description='Command-line parameters passed to spark submit.', is_required=True,
    )
    return Field(
        Shape(fields={'parameters': parameters}),
        description='Important!'
        'You can Spark submit tasks only on new clusters. '
        'In the new_cluster specification, libraries and spark_conf are not supported. Instead, '
        'use --jars and --py-files to add Java and Python libraries and use --conf to set the '
        'Spark configuration. '
        'master, deploy-mode, and executor-cores are automatically configured by Databricks; '
        'you cannot specify them in parameters. '
        'By default, the Spark submit job uses all available memory (excluding reserved memory '
        'for Databricks services). You can set --driver-memory, and --executor-memory to a '
        'smaller value to leave some room for off-heap usage. '
        'The --jars, --py-files, --files arguments support DBFS and S3 paths.',
    )


def _define_task():
    return Field(
        Selector(
            {
                'notebook_task': _define_notebook_task(),
                'spark_jar_task': _define_spark_jar_task(),
                'spark_python_task': _define_spark_python_task(),
                'spark_submit_task': _define_spark_submit_task(),
            }
        ),
        description='The task to run.',
        is_required=True,
    )


def define_databricks_submit_custom_run_config():
    fields = _define_submit_run_fields()
    fields['task'] = _define_task()
    return Field(Shape(fields=fields), description='Databricks job run configuration')


def define_databricks_submit_run_config():
    return Field(
        Shape(fields=_define_submit_run_fields()), description='Databricks job run configuration',
    )


def _define_secret_scope():
    return Field(
        String,
        description='The Databricks secret scope containing the storage secrets.',
        is_required=True,
    )


def _define_s3_storage_credentials():
    access_key_key = Field(
        String,
        description='The key of a Databricks secret containing the S3 access key ID.',
        is_required=True,
    )
    secret_key_key = Field(
        String,
        description='The key of a Databricks secret containing the S3 secret access key.',
        is_required=True,
    )
    return Field(
        Shape(
            fields={
                'secret_scope': _define_secret_scope(),
                'access_key_key': access_key_key,
                'secret_key_key': secret_key_key,
            }
        ),
        description='S3 storage secret configuration',
    )


def _define_adls2_storage_credentials():
    storage_account_name = Field(
        String,
        description='The name of the storage account used to access data.',
        is_required=True,
    )
    storage_account_key_key = Field(
        String,
        description='The key of a Databricks secret containing the storage account secret key.',
        is_required=True,
    )
    return Field(
        Shape(
            fields={
                'secret_scope': _define_secret_scope(),
                'storage_account_name': storage_account_name,
                'storage_account_key_key': storage_account_key_key,
            }
        ),
        description='ADLS2 storage secret configuration',
    )


def _define_storage_credentials():
    return Selector(
        {'s3': _define_s3_storage_credentials(), 'adls2': _define_adls2_storage_credentials()},
    )


def define_databricks_storage_config():
    return Field(
        Selector(
            {'s3': _define_s3_storage_credentials(), 'adls2': _define_adls2_storage_credentials()}
        ),
        description='Databricks storage configuration. Solids using the '
        'DatabricksPySparkStepLauncher to execute pipeline steps in Databricks MUST configure '
        'storage using this config (either S3 or ADLS2 can be used). Access credentials for the '
        'storage must be stored in Databricks secrets; this config indicates the secret scope '
        'and the secret keys used to access either S3 or ADLS2.',
        is_required=True,
    )



import base64
import time

import requests.exceptions
from databricks_api import DatabricksAPI

import dagster
from dagster import check

from .types import (
    DATABRICKS_RUN_TERMINATED_STATES,
    DatabricksRunLifeCycleState,
    DatabricksRunResultState,
)

# wait at most 24 hours by default for run execution
_DEFAULT_RUN_MAX_WAIT_TIME_SEC = 24 * 60 * 60


class DatabricksError(Exception):
    pass


class DatabricksClient:
    '''A thin wrapper over the Databricks REST API.'''

    def __init__(self, host, token):
        self.client = DatabricksAPI(host=host, token=token)

    def submit_run(self, *args, **kwargs):
        '''Submit a run directly to the 'Runs Submit' API.'''
        return self.client.jobs.submit_run(*args, **kwargs)  # pylint: disable=no-member

    def read_file(self, dbfs_path, block_size=1024 ** 2):
        '''Read a file from DBFS to a **byte string**.'''

        if dbfs_path.startswith('dbfs://'):
            dbfs_path = dbfs_path[7:]
        data = b''
        bytes_read = 0
        jdoc = self.client.dbfs.read(path=dbfs_path, length=block_size)  # pylint: disable=no-member
        data += base64.b64decode(jdoc['data'])
        while jdoc['bytes_read'] == block_size:
            bytes_read += jdoc['bytes_read']
            jdoc = self.client.dbfs.read(  # pylint: disable=no-member
                path=dbfs_path, offset=bytes_read, length=block_size
            )
            data += base64.b64decode(jdoc['data'])
        return data

    def put_file(self, file_obj, dbfs_path, overwrite=False, block_size=1024 ** 2):
        '''Upload an arbitrary large file to DBFS.

        This doesn't use the DBFS `Put` API because that endpoint is limited to 1MB.
        '''
        if dbfs_path.startswith('dbfs://'):
            dbfs_path = dbfs_path[7:]
        create_response = self.client.dbfs.create(  # pylint: disable=no-member
            path=dbfs_path, overwrite=overwrite
        )
        handle = create_response['handle']

        block = file_obj.read(block_size)
        while block:
            data = base64.b64encode(block).decode('utf-8')
            self.client.dbfs.add_block(data=data, handle=handle)  # pylint: disable=no-member
            block = file_obj.read(block_size)

        self.client.dbfs.close(handle=handle)  # pylint: disable=no-member

    def get_run_state(self, databricks_run_id):
        '''Get the state of a run by Databricks run ID (_not_ dagster run ID).

        Return a `DatabricksRunState` object. Note that the `result_state`
        attribute may be `None` if the run hasn't yet terminated.
        '''
        run = self.client.jobs.get_run(databricks_run_id)  # pylint: disable=no-member
        state = run['state']
        result_state = state.get('result_state')
        if result_state:
            result_state = DatabricksRunResultState(result_state)
        return DatabricksRunState(
            life_cycle_state=DatabricksRunLifeCycleState(state['life_cycle_state']),
            result_state=result_state,
            state_message=state['state_message'],
        )


class DatabricksRunState:
    '''Represents the state of a Databricks job run.'''

    def __init__(self, life_cycle_state, result_state, state_message):
        self.life_cycle_state = life_cycle_state
        self.result_state = result_state
        self.state_message = state_message

    def has_terminated(self):
        '''Has the job terminated?'''
        return self.life_cycle_state in DATABRICKS_RUN_TERMINATED_STATES

    def is_successful(self):
        '''Was the job successful?'''
        return self.result_state == DatabricksRunResultState.Success

    def __repr__(self):
        return str(self.__dict__)


class DatabricksJobRunner:
    '''Submits jobs created using Dagster config to Databricks, and monitors their progress.'''

    def __init__(
        self, host, token, poll_interval_sec=10, max_wait_time_sec=_DEFAULT_RUN_MAX_WAIT_TIME_SEC
    ):
        '''Args:
            host (str): Databricks host, e.g. https://uksouth.azuredatabricks.net
            token (str): Databricks token
        '''
        self.host = check.str_param(host, 'host')
        self.token = check.str_param(token, 'token')
        self.poll_interval_sec = check.numeric_param(poll_interval_sec, 'poll_interval_sec')
        self.max_wait_time_sec = check.int_param(max_wait_time_sec, 'max_wait_time_sec')

        self._client = DatabricksClient(host=self.host, token=self.token)

    @property
    def client(self):
        '''Return the underlying `DatabricksClient` object.'''
        return self._client

    def submit_run(self, run_config, task):
        '''Submit a new run using the 'Runs submit' API.'''
        existing_cluster_id = run_config['cluster'].get('existing')

        new_cluster = run_config['cluster'].get('new')

        # The Databricks API needs different keys to be present in API calls depending
        # on new/existing cluster, so we need to process the new_cluster
        # config first.
        if new_cluster:
            new_cluster = new_cluster.copy()

            nodes = new_cluster.pop('nodes')
            if 'instance_pool_id' in nodes:
                new_cluster['instance_pool_id'] = nodes['instance_pool_id']
            else:
                node_types = nodes['node_types']
                new_cluster['node_type_id'] = node_types['node_type_id']
                if 'driver_node_type_id' in node_types:
                    new_cluster['driver_node_type_id'] = node_types['driver_node_type_id']

            cluster_size = new_cluster.pop('size')
            if 'num_workers' in cluster_size:
                new_cluster['num_workers'] = cluster_size['num_workers']
            else:
                new_cluster['autoscale'] = cluster_size['autoscale']

            tags = new_cluster.get('custom_tags', [])
            tags.append({'key': '__dagster_version', 'value': dagster.__version__})
            new_cluster['custom_tags'] = tags

        check.invariant(
            existing_cluster_id is not None or new_cluster is not None,
            'Invalid value for run_config.cluster',
        )

        # We'll always need some libraries, namely dagster/dagster_databricks/dagster_pyspark,
        # since they're imported by our scripts.
        # Add them if they're not already added by users in config.
        libraries = list(run_config.get('libraries', []))
        python_libraries = {x['pypi']['package'].split('==')[0] for x in libraries if 'pypi' in x}
        for library in ["dagster", "dagster_databricks", "dagster_pyspark"]:
            if library not in python_libraries:
                libraries.append(
                    {"pypi": {"package": "{}=={}".format(library, dagster.__version__)}}
                )

        # Only one task should be able to be chosen really; make sure of that here.
        check.invariant(
            sum(
                task.get(key) is not None
                for key in [
                    'notebook_task',
                    'spark_python_task',
                    'spark_jar_task',
                    'spark_submit_task',
                ]
            )
            == 1,
            'Multiple tasks specified in Databricks run',
        )

        config = dict(
            run_name=run_config.get('run_name'),
            new_cluster=new_cluster,
            existing_cluster_id=existing_cluster_id,
            libraries=libraries,
            **task
        )
        response = self.client.submit_run(**config)
        return response['run_id']

    def wait_for_run_to_complete(self, log, databricks_run_id):
        '''Wait for a Databricks run to complete.'''
        check.int_param(databricks_run_id, 'databricks_run_id')
        log.info('Waiting for Databricks run %s to complete...' % databricks_run_id)
        start = time.time()
        while True:
            log.debug('Waiting %.1f seconds...' % self.poll_interval_sec)
            time.sleep(self.poll_interval_sec)
            run_state = self.client.get_run_state(databricks_run_id)
            if run_state.has_terminated():
                if run_state.is_successful():
                    log.info('Run %s completed successfully' % databricks_run_id)
                    return
                else:
                    error_message = 'Run %s failed with result state: %s. Message: %s' % (
                        databricks_run_id,
                        run_state.result_state,
                        run_state.state_message,
                    )
                    log.error(error_message)
                    raise DatabricksError(error_message)
            else:
                log.info('Run %s in state %s' % (databricks_run_id, run_state))
            if time.time() - start > self.max_wait_time_sec:
                raise DatabricksError(
                    'Job run {} took more than {}s to complete; failing'.format(
                        databricks_run_id, self.max_wait_time_sec
                    )
                )

    def retrieve_logs_for_run_id(self, log, databricks_run_id):
        '''Retrieve the stdout and stderr logs for a run.'''
        api_client = self.client.client
        run = api_client.jobs.get_run(databricks_run_id)  # pylint: disable=no-member
        cluster = api_client.cluster.get_cluster(  # pylint: disable=no-member
            run['cluster_instance']['cluster_id']
        )
        log_config = cluster.get('cluster_log_conf')
        if log_config is None:
            log.warn(
                'Logs not configured for cluster {cluster} used for run {run}'.format(
                    cluster=cluster['cluster_id'], run=databricks_run_id
                )
            )
            return None
        if 's3' in log_config:
            logs_prefix = log_config['s3']['destination']
            log.warn('Retrieving S3 logs not yet implemented')
            return None
        elif 'dbfs' in log_config:
            logs_prefix = log_config['dbfs']['destination']
            stdout = self.wait_for_dbfs_logs(log, logs_prefix, cluster['cluster_id'], 'stdout')
            stderr = self.wait_for_dbfs_logs(log, logs_prefix, cluster['cluster_id'], 'stderr')
            return stdout, stderr

    def wait_for_dbfs_logs(
        self, log, prefix, cluster_id, filename, waiter_delay=10, waiter_max_attempts=10
    ):
        '''Attempt up to `waiter_max_attempts` attempts to get logs from DBFS.'''
        path = '/'.join([prefix, cluster_id, 'driver', filename])
        log.info('Retrieving logs from {}'.format(path))
        num_attempts = 0
        while num_attempts <= waiter_max_attempts:
            try:
                logs = self.client.read_file(path)
                return logs.decode('utf-8')
            except requests.exceptions.HTTPError:
                num_attempts += 1
                time.sleep(waiter_delay)
        log.warn('Could not retrieve cluster logs!')



from dagster import (
    Field,
    InputDefinition,
    Int,
    Nothing,
    Output,
    OutputDefinition,
    SolidDefinition,
    String,
    check,
)

from .configs import define_databricks_submit_custom_run_config
from .databricks import DatabricksJobRunner

_START = 'start'

# wait at most 24 hours by default for run execution
_DEFAULT_RUN_MAX_WAIT_TIME_SEC = 24 * 60 * 60


class DatabricksRunJobSolidDefinition(SolidDefinition):
    '''Solid to run a Databricks job.

    The job configuration is passed entirely by config, and no code is synchronised to Databricks.
    See :py:class:`dagster_databricks.databricks_pyspark_step_launcher` to run PySpark solids within
    a Databricks job.
    '''

    def __init__(
        self,
        name,
        description=None,
        poll_interval_sec=10,
        max_wait_time_sec=_DEFAULT_RUN_MAX_WAIT_TIME_SEC,
    ):
        name = check.str_param(name, 'name')

        description = check.opt_str_param(
            description, 'description', 'Databricks run job flow solid.'
        )

        def _compute_fn(context, _):

            job_runner = DatabricksJobRunner(
                host=context.solid_config.get('databricks_host'),
                token=context.solid_config.get('databricks_host'),
                poll_interval_sec=poll_interval_sec,
                max_wait_time_sec=max_wait_time_sec,
            )
            run_config = context.solid_config['run_config'].copy()
            task = run_config.pop('task')
            databricks_run_id = job_runner.submit_run(run_config, task)
            context.log.info('waiting for Databricks job completion...')

            job_runner.wait_for_run_to_complete(context.log, databricks_run_id)
            yield Output(databricks_run_id)

        super(DatabricksRunJobSolidDefinition, self).__init__(
            name=name,
            description=description,
            input_defs=[InputDefinition(_START, Nothing)],
            output_defs=[OutputDefinition(Int)],
            compute_fn=_compute_fn,
            config_schema={
                'run_config': define_databricks_submit_custom_run_config(),
                'databricks_host': Field(String, is_required=False),
                'databricks_token': Field(String, is_required=True),
            },
        )



import os

import pytest
from dagster_aws.s3 import s3_plus_default_storage_defs, s3_resource
from dagster_databricks import databricks_pyspark_step_launcher
from dagster_pyspark import DataFrame, pyspark_resource
from pyspark.sql import Row
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from dagster import (
    InputDefinition,
    ModeDefinition,
    OutputDefinition,
    execute_pipeline,
    pipeline,
    reconstructable,
    solid,
)
from dagster.core.definitions.no_step_launcher import no_step_launcher
from dagster.seven import mock
from dagster.utils.merger import deep_merge_dicts

# TODO: uncomment when dagster-azure is merged.
# from dagster_azure.adls2 import adls2_plus_default_storage_defs, adls2_resource

S3_BUCKET = 'dagster-databricks-tests'
ADLS2_STORAGE_ACCOUNT = 'dagsterdatabrickstests'
ADLS2_CONTAINER = 'dagster-databricks-tests'


BASE_DATABRICKS_PYSPARK_STEP_LAUNCHER_CONFIG = {
    'databricks_host': os.environ.get('DATABRICKS_HOST'),
    'databricks_token': os.environ.get('DATABRICKS_TOKEN'),
    'local_pipeline_package_path': os.path.abspath(os.path.dirname(__file__)),
    'staging_prefix': '/dagster-databricks-tests',
    'run_config': {
        'cluster': {
            'new': {
                'size': {'num_workers': 1},
                'spark_version': '6.5.x-scala2.11',
                'nodes': {'node_types': {'node_type_id': 'Standard_DS3_v2'},},
            },
        },
        'libraries': [
            # TODO: uncomment when dagster-azure is merged.
            # {'pypi': {'package': 'azure-storage-file-datalake~=12.0.1'}},
            {'pypi': {'package': 'dagster-aws==0.7.13'}},
            {'pypi': {'package': 'databricks-api'}},
            {'pypi': {'package': 'pytest'}},
        ],
    },
    'storage': {
        's3': {
            'secret_scope': 'dagster-databricks-tests',
            'access_key_key': 'aws-access-key',
            'secret_key_key': 'aws-secret-key',
        }
    },
}


@solid(
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher', 'pyspark'},
)
def make_df_solid(context):
    schema = StructType([StructField('name', StringType()), StructField('age', IntegerType())])
    rows = [Row(name='John', age=19), Row(name='Jennifer', age=29), Row(name='Henry', age=50)]
    return context.resources.pyspark.spark_session.createDataFrame(rows, schema)


@solid(
    name='blah',
    description='this is a test',
    config_schema={'foo': str, 'bar': int},
    input_defs=[InputDefinition('people', DataFrame)],
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher'},
)
def filter_df_solid(_, people):
    return people.filter(people['age'] < 30)


MODE_DEFS = [
    # TODO: uncomment when dagster-azure is merged
    # ModeDefinition(
    #     'prod_adls2',
    #     resource_defs={
    #         'pyspark_step_launcher': databricks_pyspark_step_launcher,
    #         'pyspark': pyspark_resource,
    #         'adls2': adls2_resource,
    #     },
    #     system_storage_defs=adls2_plus_default_storage_defs,
    # ),
    ModeDefinition(
        'prod_s3',
        resource_defs={
            'pyspark_step_launcher': databricks_pyspark_step_launcher,
            'pyspark': pyspark_resource,
            's3': s3_resource,
        },
        system_storage_defs=s3_plus_default_storage_defs,
    ),
    ModeDefinition(
        'local',
        resource_defs={'pyspark_step_launcher': no_step_launcher, 'pyspark': pyspark_resource},
    ),
]


@pipeline(mode_defs=MODE_DEFS)
def pyspark_pipe():
    filter_df_solid(make_df_solid())


def define_pyspark_pipe():
    return pyspark_pipe


@solid(
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher', 'pyspark'},
)
def do_nothing_solid(_):
    pass


@pipeline(mode_defs=MODE_DEFS)
def do_nothing_pipe():
    do_nothing_solid()


def define_do_nothing_pipe():
    return do_nothing_pipe


def test_local():
    result = execute_pipeline(
        pipeline=reconstructable(define_pyspark_pipe),
        mode='local',
        run_config={'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}}},
    )
    assert result.success


@mock.patch('dagster_databricks.databricks.DatabricksClient.submit_run')
@mock.patch('dagster_databricks.databricks.DatabricksClient.put_file')
@mock.patch('dagster_databricks.DatabricksPySparkStepLauncher.get_step_events')
@mock.patch('dagster_databricks.databricks.DatabricksJobRunner.wait_for_run_to_complete')
def test_pyspark_databricks(mock_wait, mock_get_step_events, mock_put_file, mock_submit_run):
    result = execute_pipeline(
        pipeline=reconstructable(define_do_nothing_pipe),
        mode='prod_s3',
        run_config={
            'resources': {
                'pyspark_step_launcher': {
                    'config': deep_merge_dicts(
                        BASE_DATABRICKS_PYSPARK_STEP_LAUNCHER_CONFIG,
                        {'databricks_host': '', 'databricks_token': ''},
                    ),
                },
            },
        },
    )
    assert result.success
    assert mock_wait.call_count == 1
    assert mock_get_step_events.call_count == 1
    # TODO: uncomment this with correct value when uploaded packages are more stable
    assert mock_put_file.call_count == 3
    assert mock_submit_run.call_count == 1


@pytest.mark.skipif(
    'DATABRICKS_TEST_DO_IT_LIVE' not in os.environ,
    reason='This test is slow and requires a Databricks cluster; run only upon explicit request',
)
def test_do_it_live_databricks_s3():
    result = execute_pipeline(
        reconstructable(define_pyspark_pipe),
        mode='prod_s3',
        run_config={
            'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
            'resources': {
                'pyspark_step_launcher': {'config': BASE_DATABRICKS_PYSPARK_STEP_LAUNCHER_CONFIG},
            },
            'storage': {
                's3': {
                    'config': {
                        's3_bucket': 'dagster-databricks-tests',
                        's3_prefix': 'dagster-databricks-tests',
                    }
                }
            },
        },
    )
    assert result.success


@pytest.mark.skip(reason='Need dagster-azure to be merged')
@pytest.mark.skipif(
    'DATABRICKS_TEST_DO_IT_LIVE' not in os.environ,
    reason='This test is slow and requires a Databricks cluster; run only upon explicit request',
)
def test_do_it_live_databricks_adls2():
    config = BASE_DATABRICKS_PYSPARK_STEP_LAUNCHER_CONFIG.copy()
    config['storage'] = {
        'adls2': {
            'secret_scope': 'dagster-databricks-tests',
            'storage_account_name': ADLS2_STORAGE_ACCOUNT,
            'storage_account_key_key': 'adls2-storage-key',
        }
    }

    result = execute_pipeline(
        reconstructable(define_pyspark_pipe),
        mode='prod_adls2',
        run_config={
            'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
            'resources': {
                'pyspark_step_launcher': {'config': config},
                'adls2': {
                    'config': {
                        'storage_account': ADLS2_STORAGE_ACCOUNT,
                        'credential': {'key': os.environ.get('AZURE_STORAGE_ACCOUNT_KEY')},
                    }
                },
            },
            'storage': {
                'adls2': {
                    'config': {
                        'adls2_container': ADLS2_CONTAINER,
                        'adls2_prefix': 'dagster-databricks-tests',
                    }
                }
            },
        },
    )
    assert result.success
